﻿using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.NetworkInformation;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.Text.RegularExpressions;
using WindowsUtilities;

namespace CTNet
{
    public enum verifyFile { fileOk, exception, nullEmptyPath, nullEmptyFileName, nullWorkFolder, notDocFile, noPortFile, noZipFile };

    /*
     * Where are box selections going, or are we selecting boxes for
     * box number magnification?
     */
    public enum boxSelDest { off, boxNumMag, chkBoxCtrl, dfRetriever, document };

    /*
     * Currently using for the Refresh button and the
     * Orphan Boxes button.
     */
    public enum btnState { off, on };

    /*
     * Checkbox Control box selection message types.
     * Checkbox Control MUST have the identical types,
     * see CTClient.cs for the Checkbox Control enums.
     */
    public enum cbcBoxSelTypes { boxSets, delBoxSet, delAllBoxSets, updateSel, newCHSet, newCHSelIndex, newCBCSelIndex, setCBCState };

    // Things we tell the form about:
    public enum ctEvent { connected, connectFail, cbcBoxSetsFileNack, cbcBoxUnSel, cbcBoxSets, cbcBoxSelReq, dfrBoxSelReq, boxSelOff, shutdown, saveDoc, saveUnlockDoc, docLock, docUnLock, cbcBoxSetsReq };

    // A custom EventArgs class to pass ctClient info to main form.
    public class CtClientArgs : EventArgs
    {
        public ctEvent eventType { get; set; }
        public string eventMsg { get; set; }
    }

    public class CTClient
    {
        /*
         * We 'lock' the document meaning our connection icon turns
         * yellow, and we do not initiate messaging to Vscode. We 
         * may still receive messages which could lead to an 
         * unlocking of the document displayed by Calc Helper.
         */
        public bool vscodeLock = false;

        // Used for lockdown/unlocking.
        public btnState refreshBtnState;
        public btnState orphanBtnState;

        // Create the calc tool client event itself:
        public event EventHandler<CtClientArgs> CtClientEvent;

        // Where are box selections going?
        public boxSelDest boxSelClient;

        // We refer to this when Checkbox Control requests checkbox selections.
        public bool docHasCheckboxes = false;

        private static bool bExiting = false;
        private static bool bConnected = false;

        /*
         * Saw a problem while testing trying to connect to
         * Calc Tool Server when I knew it was not running. Got an exception in
         * the ConnectCallback() routine which caused StartClient() to hang on
         * the connectDone.WaitOne() call since there was no connectDone.Set()
         * called from ConnectCallback(). So, I am calling connectDone.Set()
         * in a finally block in ConnectCallback() along with setting this
         * exception flag and will check it in StartClient() to make sure we
         * have a 'real' connectDone.Set() signal.
         */
        public bool connException = false;

        private static int port = 0;

        private static Socket client = null;

        public VSCFileInfo vscFileInfo = null;

        // This is the arg[0] from Visual Studio Code launch.
        public string vscFile = null;

        public static string vscFileNameNoExt = null;

        // This is the file name and extension from arg[0].
        private static string workFile = null;

        // This is the directory information from arg[0].
        private static string workFolder = null;

        // This is the zip path and file name for workFile - ext is .zip.
        //private static string zipFile = null;
        public string zipFile = null;

        // Location of Calc Tool Server port number file so client tool can connect.
        private static string portFile = null;

        private static IntPtr _hWnd = IntPtr.Zero;
        
        private static CTMessage gMsg = new CTMessage();

        // Thread signal. Probably will not use some of these..... 
        private static ManualResetEvent allDone = new ManualResetEvent(false);
        private static ManualResetEvent msgDone = new ManualResetEvent(false);

        private static ManualResetEvent sendMsgDone = new ManualResetEvent(false);
        private static ManualResetEvent receiveMsgDone = new ManualResetEvent(false);

        private static ManualResetEvent sendDone = new ManualResetEvent(false);
        private static ManualResetEvent connectDone = new ManualResetEvent(false);
        private static ManualResetEvent connectAckDone = new ManualResetEvent(false);

        public CTClient(VSCFileInfo vscodeFileInfo)
        {
            boxSelClient = boxSelDest.off;

            vscFileInfo = vscodeFileInfo;

            SetFileInfo(true);
        }

        private void SetFileInfo(bool doPortFile)
        {
            /*
             * I think this is called 'normalizing' the file path,
             * so c:\blah\blah and c:/blah/blah are comparable
             * to each other. Needed when we get a file 
             * open/close/save message from Vscode.
             */
            vscFile = Path.GetFullPath(vscFileInfo.fileTmp).ToLower();

            vscFileNameNoExt = vscFileInfo.fileNameNoExtTmp;
            workFolder = vscFileInfo.workFolderTmp;
            workFile = vscFileInfo.workFileTmp;
            zipFile = vscFileInfo.zipFileTmp;

            if (doPortFile == true)
            {
                portFile = vscFileInfo.portFileTmp;
            }
        }

        // Trigger the calc tool client event:
        protected virtual void OnCtClientEvent(CtClientArgs e)
        {
            if (CtClientEvent != null)
            {
                CtClientEvent(this, e);
            }
        }

        /*
         * Tell a tool that is currently receiving box selections
         * from Calc Helper that it will no longer receive them.
         * 
         * Called From:
         *
         *  ProcessChkBoxCtrlMode()
         *  ProcessDFRetMode()
         * 
         * Arguments:
         * 
         *  boxSelDest - The tool that is currently receiving box selections.
         */
        public void BoxSelOff(boxSelDest boxSelDest)
        {
            if (client != null && client.Connected == true)
            {
                bool sendMsg = true;
                CTMessage msg = new CTMessage();

                switch (boxSelDest)
                {
                    case (boxSelDest.chkBoxCtrl):
                        msg.Command = "CHBoxSelOffCBC";
                        break;

                    case (boxSelDest.dfRetriever):
                        msg.Command = "CHBoxSelOffDFR";
                        break;

                    default:
                        sendMsg = false;
                        break;
                }

                if (sendMsg == true)
                {
                    SendAsync(msg);
                }
            }
        }

        public bool IsConnected()
        {
            bool result = false;

            if (client != null && client.Connected == true)
            {
                result = true;
            }

            return result;
        }

        public void StopClient()
        {
            bExiting = true;
            receiveMsgDone.Set();

            if (client != null && client.Connected == true)
            {
                try
                {
                    client.Shutdown(SocketShutdown.Both);
                    client.Close();
                }
                catch
                {

                }
            }
        }

        public void StartClient()
        {
            CtClientArgs ctClientArgs = new CtClientArgs();

            // Connect to Visual Studio Code server.
            if (ReadPortNumber() == false)
            {
                MessageBox.Show("  Could not read .calctool port number file.\n\n  Calc Helper launch will be aborted.", "Calc Tool Server Connection Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                ctClientArgs.eventType = ctEvent.connectFail;
            }
            else
            {
                try
                {
                    // Establish the remote endpoint for the socket.  
                    IPAddress ipAddress = IPAddress.Loopback;
                    IPEndPoint remoteEP = new IPEndPoint(ipAddress, port);

                    // Create a TCP/IP socket.  
                    client = new Socket(ipAddress.AddressFamily, SocketType.Stream, ProtocolType.Tcp);

                    // Connect to the remote endpoint.  
                    client.BeginConnect(remoteEP, new AsyncCallback(ConnectCallback), client);

                    connectDone.WaitOne();

                    // Debug:
                    // Added this if (..) { .. } else { ... } to see if we can abort and close
                    // this application if ConnectCallback() has an exception.
                    if (connException)
                    {
                        //MessageBox.Show("  We have a connException!", "Calc Helper-StartClient():", MessageBoxButtons.OK, MessageBoxIcon.Information);
                        ctClientArgs.eventType = ctEvent.connectFail;
                    }
                    else
                    {
                        // Send connect message to Calc Tool Server.
                        SendConnectMsg();

                        if (sendMsgDone.WaitOne(10000) == true)
                        {
                            // Now we expect the Connection Acknowledgment.
                            //MessageBox.Show("  Waiting for connection acknowledgment from Calc Tool Server.", "Calc Helper-StartClient():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                            receiveMsgDone.Reset();

                            ReceiveAsync(client);

                            if (receiveMsgDone.WaitOne(30000) == true)
                            {
                                // Signal main form that we are connected to Calc Tool Server.
                                ctClientArgs.eventType = ctEvent.connected;

                                //MessageBox.Show("  Calc Tool Server has acknowledged our connection!.", "Calc Helper-StartClient():", MessageBoxButtons.OK, MessageBoxIcon.Information);
                            }
                            else
                            {
                                MessageBox.Show("  Timed out connecting to Calc Tool Server\n  while executing receiveMsgDone.WaitOne(30000)\n\n  Calc Helper launch will be aborted.", "Calc Tool Server Connection Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                                ctClientArgs.eventType = ctEvent.connectFail;
                            }
                        }
                        else
                        {
                            MessageBox.Show("  Timed out connecting to Calc Tool Server\n  while executing sendMsgDone.WaitOne(10000)\n\n  Calc Helper launch will be aborted.", "Calc Tool Server Connection Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                            ctClientArgs.eventType = ctEvent.connectFail;
                        }
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString(), "Calc Tool Server Connection Error:", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    ctClientArgs.eventType = ctEvent.connectFail;
                }
            }

            OnCtClientEvent(ctClientArgs);
        }

        public void SetVscodeFocus()
        {
            string folder = workFolder.Substring(workFolder.LastIndexOf('\\') + 1);
            string wndwTitle = string.Format("{0} - {1} - Visual Studio Code", workFile, folder);

            //MessageBox.Show(string.Format("Looking for: {0}", wndwTitle), "SetVscodeFocus():", MessageBoxButtons.OK, MessageBoxIcon.Information);

            try
            {
                IEnumerable<IntPtr> hWnd = WinUtils.FindWindowsWithText(wndwTitle);

                if (hWnd.Count() > 0)
                {
                    //MessageBox.Show(string.Format("hWnd.Count() = {0}", hWnd.Count()), "Entered: if (hWnd.Count() > 0)", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    WinUtils.SetForegroundWindow(hWnd.ElementAt(0));
                }
                else
                {
                    MessageBox.Show(string.Format("  Could not set Visual Studio Code Focus for:\n\n  {0}\n\n  You will need to manually mouse click the\n  Visual Studio Code window to give it focus.", wndwTitle), "SetVscodeFocus() Error:", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "CTClient-SetVscodeFocus():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * Sends an insert snippet command and the snippet name to
         * the Calc Tool server. The snippet will be inserted at the
         * current cursor location in Vscode.
         * 
         * Called From:
         * 
         * snipInsertBtn_Click()
         */
        public void InsertSnippet(string snipName)
        {
            if (client != null && client.Connected == true)
            {
                CTMessage msg = new CTMessage();

                msg.Command = "InsertSnippet";
                msg.Data = snipName;

                SetVscodeFocus();

                SendAsync(msg);
            }
        }

        /*
         * Sends box selections to the current receiver of
         * selections.
         * 
         * Called From:
         * 
         * ProcessRightMouseClick()
         * SendOffLeftBoxSel()
         * indBoxCalcToVarBtn_Click()
         */
        public void SendBoxSelections(string boxSel)
        {
            if (client != null && client.Connected == true)
            {
                CTMessage msg = new CTMessage();

                // Who are we sending these boxes to?
                switch (boxSelClient)
                {
                    case (boxSelDest.chkBoxCtrl):

                        msg.Command = "BoxSelCBC";
                        msg.FileName = vscFile;
                        msg.Data = boxSel;
                        break;

                    case (boxSelDest.dfRetriever):

                        msg.Command = "BoxSelDFR";
                        msg.FileName = vscFile;
                        msg.Data = boxSel;
                        break;

                    case (boxSelDest.document):

                        msg.Command = "BoxSelDOC";

                        // Server needs the clipboard for this command.
                        Clipboard.SetText(boxSel, TextDataFormat.Text);
                        SetVscodeFocus();
                        break;
                }

                SendAsync(msg);
            }
        }

        /*
         * Sends a Grouping Key from the Tool Belt's Box
         * Selector to Vscode via the SendKeys command. I
         * really do not completely get how this works but
         * it works, just as if you manually pressed the
         * keyboard key.
         * 
         * We do not need to actually send a command to the
         * Calc Tool Server. We only need to set the focus
         * to Vscode. We also can only do this command when
         * we are in Document box selection mode which the
         * caller has verified - if not the caller it would
         * be the caller's caller.
         * 
         * Called From:
         * 
         *      grpBracketBtn_Click()
         *      grpDollCurlyBraceBtn_Click()
         *      grpCurlyBraceBtn_Click()
         *      grpParenBtn_Click()
         *      
         * Arguments:
         * 
         *      grpKey - The key to send, [ or ( or { or
         *          $ followed by {, actually 2 keys.
        */
        public void SendGroupingKey(string grpKey)
        {
            bool sendKey = true;
            string sendStr = "";

            switch (grpKey)
            {
                case ("${"):
                    sendStr = "{$}{{}";
                    break;

                case ("{"):

                    sendStr = "{{}";
                    break;

                case ("("):

                    sendStr = "{(}";
                    break;

                case ("["):

                    sendStr = "{[}";
                    break;

                default:
                    sendKey = false;
                    break;
            }

            if (sendKey == true)
            {
                SetVscodeFocus();
                SendKeys.Send(sendStr);
            }
        }

        /*
         * Send a command to the Calc Tool Server to go to code
         * in Vscode. This can relate to a box being controlled
         * like a Calculation, Dataflow, Bluefield Control,
         * Checkbox Control, or a comment line, a method
         * declaration line, an array, etc..
         * 
         * We pass our codeStr contents to the server via the
         * clipboard as this is the only way I could find to the 
         * search string into the find widget [via pasting from
         * the clipboard].
         * 
         * Called From:
         * 
         *      ctrlArrayList_Click()
         *      docFcnList_Click()
         *      comMarkerList_Click()
         *      orphanBoxList_Click()
         *      GotoControlCode()
         *      
         * Arguments:
         * 
         *      closeFindWidget - 'false' means we want to keep
         *          the find widget open in Vscode. This is done
         *          for control code like Dataflow as an example.
         *          'true' obviuosly means we close the find as
         *          we only expect one occurrance of what we are
         *          tring to find.
         * 
         *          NOTE: THIS MAY NEED TWEAKING IF I CAN FIGURE 
         *          OUT HOW TO SET THE VSCODE'S FIND WIDGET TO
         *          SEARCH USING REGULAR EXPRESSION OR TO UNSET
         *          THAT FOR SEARCH STRINGS THAT DO NOT CONTAIN
         *          REGEXES.
         *          
         *          MAY END UP HAVING ANOTHER GOTO FUNCTION IN
         *          ADDITION TO THIS ONE, TWEAKED OR NOT.
         *          
         *      scrollFirstLine - 'true' means the Calc Tool Server
         *          will first scroll to the first line in the code
         *          file. Using this for Checkbox Control goto's as
         *          I want to close the find widget, and I want to
         *          start at the beginning of the file where these
         *          checkboxes are kept.
         *          
         *      codeStr - Our object of interest which we are trying
         *          to find.
        */
        public void GotoCode(bool closeFindWidget, bool scrollFirstLine, string codeStr)
        {
            if (client != null && client.Connected == true)
            {
                CTMessage msg = new CTMessage();

                if (closeFindWidget == true)
                {
                    msg.Command = "GotoCodeCloseFind";
                    msg.Data = (scrollFirstLine == true ? "ToFirstLine" : "");
                }
                else
                {
                    msg.Command = "GotoCode";
                }

                Clipboard.SetText(codeStr, TextDataFormat.Text);
                SetVscodeFocus();

                SendAsync(msg);
            }
        }

        // From Joe's SBClient:
        public void SendConnectMsg()
        {
            CTMessage msg = new CTMessage();
            msg.Command = "Connect";

            SendAsync(msg);
        }

        // From Joe's SBClient:
        public void SendAsync(CTMessage msg)
        {
            if (client != null && client.Connected)
            {
                msg.Id = "CH";
                string json = JsonConvert.SerializeObject(msg);

                byte[] bytes = Encoding.ASCII.GetBytes(json);

                byte[] head = BitConverter.GetBytes(bytes.Length);
                byte[] dmsg = new byte[bytes.Length + head.Length];
                head.CopyTo(dmsg, 0);
                bytes.CopyTo(dmsg, head.Length);

                // Begin sending the data to the remote device.  
                client.BeginSend(dmsg, 0, dmsg.Length, 0, new AsyncCallback(SendCallback), client);
            }
        }

        // From Joe's SBClient:
        public void SendCallback(IAsyncResult ar)
        {
            try
            {
                // Retrieve the socket from the state object.  
                Socket client = (Socket)ar.AsyncState;

                // Complete sending the data to the remote device.  
                int bytesSent = client.EndSend(ar);
                Console.WriteLine("Sent {0} bytes to server.", bytesSent);

                // Signal that all bytes have been sent.  
                sendMsgDone.Set();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }

        public bool ReadPortNumber()
        {
            bool ret = false;

            try
            {
                string text = File.ReadAllText(portFile, Encoding.UTF8);
                ret = int.TryParse(text, out port);
            }
            catch (Exception ex)
            {
                ret = false;
                MessageBox.Show(ex.ToString(), "CTClient-ReadPortNumber():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            return ret;
        }

        public void ConnectCallback(IAsyncResult ar)
        {
            CtClientArgs ctClientArgs = new CtClientArgs();

            try
            {
                // Retrieve the socket from the state object.  
                Socket client = (Socket)ar.AsyncState;

                // Complete the connection.  
                client.EndConnect(ar);

                //MessageBox.Show(string.Format("Socket connected to {0}", client.RemoteEndPoint.ToString()), "Connected:", MessageBoxButtons.OK, MessageBoxIcon.Information);

                // Signal that the connection has been made.  
                connectDone.Set();
            }
            catch (Exception ex)
            {
                /*
                 * Discovered tricky error condition while purposefully launching Calc Helper
                 * when the Calc Tool Server is not running. The form is blocked by the connectDone.WaitOne()
                 * call in StartClient(), so doing this and calling connectDone.Set() to kickstart execution
                 * but will be checking for connException to see if it was this situation.
                 */
                connException = true;

                MessageBox.Show(ex.ToString(), "CTClient-ConnectCallback() - Calc Helper Launch Will Be Aborted:", MessageBoxButtons.OK, MessageBoxIcon.Error);

                connectDone.Set();
            }
            // Debug - Start:
            //finally
            //{
            //    //connException = true;
            //    //connectDone.Set();
            //}
            // Debug - Stop.
        }

        public void ReceiveAsync(Socket skt)
        {
            if (skt != null && skt.Connected)
            {
                InitStateObject isa = new InitStateObject();
                isa.workSocket = skt;

                try
                {
                    skt.BeginReceive(isa.LenBuffer, 0, isa.LenBuffer.Length, SocketFlags.None, new AsyncCallback(InitAsyncReadCallback), isa);
                }
                catch
                {

                }
            }
        }

        public void InitAsyncReadCallback(IAsyncResult ar)
        {
            try
            {
                InitStateObject isa = ar.AsyncState as InitStateObject;

                Socket client = isa.workSocket;

                int rcvd = client.EndReceive(ar);

                if (rcvd != 4)
                    throw new Exception();

                int to_recv = BitConverter.ToInt32(isa.LenBuffer, 0);

                // Create the state object.  
                StateObject state = new StateObject(to_recv);
                state.workSocket = client;
                state.toReceive = to_recv;

                client.BeginReceive(state.buffer, 0, state.buffer.Length, SocketFlags.None, new AsyncCallback(AsyncReadCallback), state);
            }
            catch
            {

            }
        }

        public void AsyncReadCallback(IAsyncResult ar)
        {
            StateObject state = ar.AsyncState as StateObject;
            Socket client = state.workSocket;

            int rcvd = client.EndReceive(ar);

            if (rcvd <= 0)
                return;

            state.BufStream.Write(state.buffer, 0, rcvd);

            state.toReceive -= rcvd;

            if (state.toReceive > 0)
            {
                Array.Clear(state.buffer, 0, state.buffer.Length);
                client.BeginReceive(state.buffer, 0, state.buffer.Length, SocketFlags.None, new AsyncCallback(AsyncReadCallback), state);
                return;
            }

            state.BufStream.Position = 0;

            string message = Encoding.ASCII.GetString(state.BufStream.ToArray());

            state.BufStream.Close();

            state.message = JsonConvert.DeserializeObject<CTMessage>(message);

            //MessageBox.Show(String.Format("Id: {0}, Command: {1}, FileName: {2}, Status: {3}", state.message.Id, state.message.Command, state.message.FileName, state.message.Status), "CTClient-AsyncReadCallback():", MessageBoxButtons.OK, MessageBoxIcon.Information);

            ProcessMessage(client, state.message);

            // 06/04/19 Just added this..so far CBC request for boxes times out...
            receiveMsgDone.Reset();

            ReceiveAsync(client);
        }

        public void ProcessMessage(Socket client, CTMessage message)
        {
            string msgFile = "";
            CtClientArgs ctClientArgs;
            CTMessage msg;

            //MessageBox.Show("Message: " + message.Command, "CTClient-ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

            switch (message.Command)
            {
                case ("ConnectAck"):

                    if (message.Id.StartsWith("VSC") == true)
                    {
                        receiveMsgDone.Set();

                        //MessageBox.Show("Calc Tool Server has acknowledged the connection!", "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);
                        
                        bConnected = true;
                    }

                    break;

                case ("CBCBoxSets"):

                    //MessageBox.Show("CBC sent current box sets", "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    ctClientArgs = new CtClientArgs();
                    //msgFile = Path.GetFullPath(message.FileName).ToLower();

                    /*
                     * This has been refitted to convey that the Checkbox Control
                     * user has either selected a new setList or removed the selected
                     * box set from the list, or they want to start a new set, or
                     * they are sending the current checkbox sets as checkbox
                     * control is closing.
                     * 
                     * Checks Performed:
                     * 
                     * 1) Is Calc Helper currently sending checkbox selections to Checkbox
                     *    Control? Just a sanity check.
                     * 
                     * If the above check fails we will ignore the message and show a Calc Helper
                     * message to the user.
                     */
                    if (boxSelClient == boxSelDest.chkBoxCtrl)
                    {
                        //if (msgFile == vscFile)
                        //{
                            ctClientArgs.eventType = ctEvent.cbcBoxSets;
                            ctClientArgs.eventMsg = message.Data;
                        //}
                        //else
                        //{
                            //ctClientArgs.eventType = ctEvent.cbcBoxSetsFileNack;

                            //ctClientArgs.eventMsg = String.Format("Checkbox Control Checkbox Sets Error:\n\nCheckbox Control just sent the current Checkbox Set list contents for a different file.\n\nThe file being displayed in Calc Helper:\n\n{0}\n\nIs not the same file that Checkbox Control is linked to:\n\n{1}\n\nCalc Helper will ignore the box sets.", vscFile, msgFile);
                        //}
                    }
                    else
                    {
                        ctClientArgs.eventType = ctEvent.cbcBoxSetsFileNack;

                        ctClientArgs.eventMsg = "Checkbox Control Checkbox Index Error:\n\nCheckbox Control just sent a 'Check Box Set' list index message to Calc Helper but Calc Helper's 'Box Selection Mode' is not 'Checkbox Control Tool'.\n\nCalc Helper will ignore the message.";
                    }

                    OnCtClientEvent(ctClientArgs);

                    break;

                case ("CBCBoxUnSel"):

                    //MessageBox.Show("CBC sent box(es) to unselect", "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    /*
                     * Not using this - will leave it in but commented out.
                     * 
                     * Checkbox Control has just sent one or more boxes to unselect. Not going
                     * nuts over the file name comparison thing, one reason we could be receiving
                     * this message is because the file names between Calc Helper and Checkbox
                     * Control ARE different. Another could be file names ARE the same, but a
                     * checkbox is already present in the Checkbox Control box list for a set.
                     */
                    //ctClientArgs = new CtClientArgs();

                    //ctClientArgs.eventType = ctEvent.cbcBoxUnSel;
                    //ctClientArgs.eventMsg = message.Data;
                    //OnCtClientEvent(ctClientArgs);

                    break;

                case ("CBCBoxSel"):

                    //MessageBox.Show("CBC wants some boxes/box sets", "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    
                    /*
                     * Calc Helper will get a second request for box selections
                     * from CBC after CBC receives the selection ACK message.
                     * The file name will be the request for the check box sets.
                     */
                    if (message.FileName == "ChkBoxSetsReq")
                    {
                        ctClientArgs = new CtClientArgs();
                        ctClientArgs.eventType = ctEvent.cbcBoxSetsReq;
                        OnCtClientEvent(ctClientArgs);
                    }
                    else
                    {
                        /*
                         * Checkbox Control is requesting check box selections.
                         * We must have checkboxes on the document and we must
                         * make sure both tools are connected to the same file.
                         */
                        msg = new CTMessage();

                        /*
                         * We use this command to indicate one of these and tell CBC
                         * the bad news. This is the best I could do given my knowledge
                         * of server mechanics, probably a better way to do all of this,
                         * plus changing the calc tool server extension is not possible.
                         */
                        if ((docHasCheckboxes == false) || (vscodeLock == true))
                        {
                            msg.Command = "CHBoxSelOffCBC";
                        }
                        else
                        {
                            if (Path.GetFullPath(message.FileName).ToLower() == vscFile)
                            {
                                msg.Command = "CBCBoxSelAck";

                                /*
                                 * Being anal here, we should not already be in this box
                                 * selection mode.
                                 */
                                if (boxSelClient != boxSelDest.chkBoxCtrl)
                                {
                                    ctClientArgs = new CtClientArgs();
                                    ctClientArgs.eventType = ctEvent.cbcBoxSelReq;
                                    OnCtClientEvent(ctClientArgs);
                                }
                            }
                            else
                            {
                                //MessageBox.Show(string.Format("CBC wants some boxes but file names differ\n  message.FileName = {0}\n  CH file name = {1}", message.FileName, vscFileNameNoExt), "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                                msg.Command = "CBCBoxSelFileNack";
                                msg.FileName = vscFile;
                            }
                        }

                        SendAsync(msg);
                    }

                    break;

                case ("DFRBoxSel"):

                    //MessageBox.Show("DFR wants some boxes", "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    /*
                     * Dataflow Retriever is requesting box selections.
                     * We must make sure both tools are connected to the same file.
                     */
                    msg = new CTMessage();

                    if (Path.GetFullPath(message.FileName).ToLower() == vscFile)
                    {
                        msg.Command = "DFRBoxSelAck";

                        /*
                         * Being anal here, we should not already be in this box
                         * selection mode.
                         */
                        if (boxSelClient != boxSelDest.dfRetriever)
                        {
                            ctClientArgs = new CtClientArgs();
                            ctClientArgs.eventType = ctEvent.dfrBoxSelReq;
                            OnCtClientEvent(ctClientArgs);
                        }
                    }
                    else
                    {
                        //MessageBox.Show(string.Format("DFR wants some boxes but file names differ\n  message.FileName = {0}\n  CH file name = {1}", message.FileName, vscFileNameNoExt), "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                        msg.Command = "DFRBoxSelFileNack";
                        msg.FileName = vscFile;
                    }

                    SendAsync(msg);

                    break;

                case ("CBCBoxSelOff"):

                    /*
                     * Checkbox Control no longer wants check box selections.
                     * Tell the form about this.
                     */
                    if (boxSelClient == boxSelDest.chkBoxCtrl)
                    {
                        ctClientArgs = new CtClientArgs();
                        ctClientArgs.eventType = ctEvent.boxSelOff;
                        OnCtClientEvent(ctClientArgs);
                    }

                    break;

                case ("DFRBoxSelOff"):

                    /*
                     * Dataflow Retriever no longer wants box selections.
                     * Tell the form about this.
                     */
                    if (boxSelClient == boxSelDest.dfRetriever)
                    {
                        ctClientArgs = new CtClientArgs();
                        ctClientArgs.eventType = ctEvent.boxSelOff;
                        OnCtClientEvent(ctClientArgs);
                    }

                    break;

                case ("Disconnect-CBC"):

                    //MessageBox.Show("CBC has disconnected from tool server.", "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    /*
                     * Checkbox Control has disconnected from the Calc Tool Server.
                     * Tell the form about this if selecting boxes for Checkbox Control.
                     */
                    if (boxSelClient == boxSelDest.chkBoxCtrl)
                    {
                        ctClientArgs = new CtClientArgs();
                        ctClientArgs.eventType = ctEvent.boxSelOff;
                        OnCtClientEvent(ctClientArgs);
                    }

                    break;

                case ("Disconnect-DFR"):

                    //MessageBox.Show("DFR has disconnected from tool server.", "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    /*
                     * Dataflow Retriever has disconnected from the Calc Tool Server.
                     * Tell the form about this if selecting boxes for Dataflow Retriever.
                     */
                    if (boxSelClient == boxSelDest.dfRetriever)
                    {
                        ctClientArgs = new CtClientArgs();
                        ctClientArgs.eventType = ctEvent.boxSelOff;
                        OnCtClientEvent(ctClientArgs);
                    }

                    break;

                case ("OpenDoc"):

                    msgFile = Path.GetFullPath(message.FileName).ToLower();

                    if (msgFile == vscFile)
                    {
                        if (vscodeLock == true)
                        {
                            /*
                             * Somehow our editor is back 'on' our file, so unlock our
                             * messaging to the editor.
                             */
                            vscodeLock = false;

                            ctClientArgs = new CtClientArgs();

                            ctClientArgs.eventType = ctEvent.docUnLock;
                            OnCtClientEvent(ctClientArgs);
                        }
                    }
                    else
                    {
                        /*
                         * A file has been opened in vscode that is not the
                         * Calc Helper linked file. Lock Calc Helper.
                         */
                        if (vscodeLock == false)
                        {
                            string boxSelClientMsg = "";

                            switch (boxSelClient)
                            {
                                case (boxSelDest.chkBoxCtrl):

                                    boxSelClientMsg = "\n\nCheckbox Control will no longer be receiving checkbox\nselections from Calc Helper.";
                                    break;

                                case (boxSelDest.dfRetriever):

                                    boxSelClientMsg = "\n\nDataflow Retriever will no longer be receiving\ncheckbox selections from Calc Helper.";
                                    break;
                            }

                            vscodeLock = true;

                            ctClientArgs = new CtClientArgs();

                            ctClientArgs.eventType = ctEvent.docLock;

                            ctClientArgs.eventMsg = String.Format("Editor opened a different file.\n\nThe file being displayed in Calc Helper:\n\n{0}\n\nIs no longer the active file in Visual Studio Code:\n\n{1}\n\nCalc Helper is suspending interaction with Visual Studio Code.{2}", vscFile, msgFile, boxSelClientMsg);

                            OnCtClientEvent(ctClientArgs);
                        }
                    }

                    break;

                case ("CloseDoc"):

                    if (Path.GetFullPath(message.FileName).ToLower() == vscFile)
                    {
                        // Trouble in paradise....their closing our file!
                        if (vscodeLock == false)
                        {
                            string boxSelClientMsg = "";

                            switch (boxSelClient)
                            {
                                case (boxSelDest.chkBoxCtrl):

                                    boxSelClientMsg = "\n\nCheckbox Control will no longer be receiving checkbox\nselections from Calc Helper.";
                                    break;

                                case (boxSelDest.dfRetriever):

                                    boxSelClientMsg = "\n\nDataflow Retriever will no longer be receiving\ncheckbox selections from Calc Helper.";
                                    break;
                            }

                            vscodeLock = true;

                            ctClientArgs = new CtClientArgs();

                            ctClientArgs.eventType = ctEvent.docLock;

                            ctClientArgs.eventMsg = String.Format("Calc Helper File Closed in Vscode.\n\nThe file being displayed in Calc Helper:\n\n{0}\n\nHas just been closed in Visual Studio Code.\n\nCalc Helper is suspending interaction with Visual Studio Code.{1}", vscFile, boxSelClientMsg);

                            OnCtClientEvent(ctClientArgs);
                        }
                    }

                    break;

                case ("ActiveTextEditorDelta"):

                    msgFile = Path.GetFullPath(message.FileName).ToLower();

                    if (msgFile == vscFile)
                    {
                        if (vscodeLock == true)
                        {
                            vscodeLock = false;

                            ctClientArgs = new CtClientArgs();

                            ctClientArgs.eventType = ctEvent.docUnLock;
                            OnCtClientEvent(ctClientArgs);
                        }
                    }
                    else
                    {
                        // Trouble in paradise....new vscode editor file tab.
                        if (vscodeLock == false)
                        {
                            string boxSelClientMsg = "";

                            switch (boxSelClient)
                            {
                                case (boxSelDest.chkBoxCtrl):

                                    boxSelClientMsg = "\n\nCheckbox Control will no longer be receiving checkbox\nselections from Calc Helper.";
                                    break;

                                case (boxSelDest.dfRetriever):

                                    boxSelClientMsg = "\n\nDataflow Retriever will no longer be receiving\ncheckbox selections from Calc Helper.";
                                    break;
                            }

                            vscodeLock = true;

                            ctClientArgs = new CtClientArgs();

                            ctClientArgs.eventType = ctEvent.docLock;

                            ctClientArgs.eventMsg = String.Format("Editor switched to a different file.\n\nThe file being displayed in Calc Helper:\n\n{0}\n\nIs no longer the active file in Visual Studio Code:\n\n{1}\n\nCalc Helper is suspending interaction with Visual Studio Code.{2}", vscFile, msgFile, boxSelClientMsg);

                            OnCtClientEvent(ctClientArgs);
                        }
                    }

                    break;

                case ("Disconnecting"):

                    if (message.Id.StartsWith("VSC") == true)
                    {
                        // Vscode is going down, bye bye time...

                        ctClientArgs = new CtClientArgs();
                        
                        ctClientArgs.eventType = ctEvent.shutdown;
                        OnCtClientEvent(ctClientArgs);
                    }

                    break;

                case ("SaveDoc"):

                    //MessageBox.Show(String.Format("Text Document Saved In Visual Studio Code:\n\n  {0}\n\nvscFile: {1}", Path.GetFullPath(message.FileName).ToLower(), vscFile), "CH ProcessMessage():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                    msgFile = Path.GetFullPath(message.FileName).ToLower();

                    if (msgFile == vscFile)
                    {
                        ctClientArgs = new CtClientArgs();

                        /*
                         * Can't 'see' how this would happen, but....
                         */
                        if (vscodeLock == true)
                        {
                            vscodeLock = false;
                            ctClientArgs.eventType = ctEvent.saveUnlockDoc;
                        }
                        else
                        {
                            ctClientArgs.eventType = ctEvent.saveDoc;
                        }

                        OnCtClientEvent(ctClientArgs);
                    }
                    else
                    {
                        if (vscodeLock == false)
                        {
                            // Trouble in paradise....Not sure how this would happen either..

                            string boxSelClientMsg = "";

                            switch (boxSelClient)
                            {
                                case (boxSelDest.chkBoxCtrl):

                                    boxSelClientMsg = "\n\nCheckbox Control will no longer be receiving checkbox\nselections from Calc Helper.";
                                    break;

                                case (boxSelDest.dfRetriever):

                                    boxSelClientMsg = "\n\nDataflow Retriever will no longer be receiving\ncheckbox selections from Calc Helper.";
                                    break;
                            }

                            vscodeLock = true;

                            ctClientArgs = new CtClientArgs();

                            ctClientArgs.eventMsg = String.Format("Mismatching Files.\n\nText Document Saved In Visual Studio Code:\n\n{0}\n\nDoes not match Calc Helper file:\n\n{1}\n\nCalc Helper is suspending interaction with Visual Studio Code.{2}", msgFile, vscFile, boxSelClientMsg);

                            ctClientArgs.eventType = ctEvent.docLock;
                            OnCtClientEvent(ctClientArgs);
                        }
                    }

                    break;
            }
        }
    }

    public class VSCFileInfo
    {
        public string fileTmp = null;
        public string fileNameNoExtTmp = null;
        public string workFolderTmp = null;
        public string workFileTmp = null;
        public string zipFileTmp = null;
        public string zipFileNameTmp = null;
        public string portFileTmp = null;

        public verifyFile verifyResult;

        public VSCFileInfo()
        {
            portFileTmp = "C:\\temp\\CalcHelper\\.calctools";
        }

        /*
         * Calc Helper has been launched from Visual Studio with this file
         * argument - we must verify before we can continue with the launch
         * process.
         */
        public void VerifyFileInfo(string vscItemPath)
        {
            //string zipFileName;

            verifyResult = verifyFile.fileOk;

            try
            {
                if (string.IsNullOrEmpty(vscItemPath))
                {
                    verifyResult = verifyFile.nullEmptyPath;
                }
                else
                {
                    fileTmp = vscItemPath;
                    workFileTmp = Path.GetFileName(vscItemPath);

                    if (string.IsNullOrEmpty(workFileTmp))
                    {
                        verifyResult = verifyFile.nullEmptyFileName;
                    }
                    else 
                    {
                        if (Regex.Match(workFileTmp.ToLower(), @"^[a-z]{4}\d{4}\.js$").Success)
                        {
                            workFolderTmp = GetFolder(vscItemPath);

                            if (workFolderTmp != null)
                            {
                                /*
                                 * Verify existance of .calctool port number file:
                                 * 
                                 * portFileTmp has been initialized in VSCFileInfo's
                                 * constructor.
                                 */
                                if (!File.Exists(portFileTmp))
                                {
                                    verifyResult = verifyFile.noPortFile;
                                    return;
                                }

                                /*
                                 * It is logical to first check where Calc Helper launched from
                                 * when searching for the document file's zip file, though I know
                                 * [at least as of now 12/2019] it will not be there.
                                 */
                                fileNameNoExtTmp = Path.GetFileNameWithoutExtension(workFileTmp).ToUpper();
                                zipFileNameTmp = fileNameNoExtTmp + ".zip";
                                zipFileTmp = Path.Combine(workFolderTmp, zipFileNameTmp);

                                if (!File.Exists(zipFileTmp))
                                {
                                    /*
                                     * Ok, so no zip file found - next guess will be our work
                                     * folder on our C: drive, again this is where I expect it to be.
                                     */
                                    zipFileTmp = Path.Combine(CalcHelper.Form1.pdfCurDir, zipFileNameTmp);

                                    if (!File.Exists(zipFileTmp))
                                    {
                                        /*
                                         * Hmmm, this may be a situation where the zip file is
                                         * in a training folder of some kind/location, so caller
                                         * will ask the user where the document's .zip file is.
                                         */
                                        verifyResult = verifyFile.noZipFile;
                                    }
                                }
                            }
                            else
                            {
                                verifyResult = verifyFile.nullWorkFolder;
                            }
                        }
                        else
                        {
                            verifyResult = verifyFile.notDocFile;
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                verifyResult = verifyFile.exception;
                MessageBox.Show(ex.ToString(), "CTClient-VerifyFileInfo():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        public string GetFolder(string vscItemPath)
        {
            string folder = null;

            // Caller has ensured that vscEditorFile is not null.

            folder = Path.GetDirectoryName(vscItemPath);

            if (folder == null)
            {
                // Root path or the empty string.
                folder = Path.GetPathRoot(vscItemPath);
            }

            return folder;
        }
    }

    public class StateObject
    {
        public Socket workSocket = null;
        public byte[] buffer; // = new byte[BufferSize];

        public int toReceive;
        public MemoryStream BufStream;
        public CTMessage message;

        public StateObject(int toRec)
        {
            toReceive = toRec;
            BufStream = new MemoryStream(toRec);
            buffer = new byte[toRec];
        }
    }

    public class InitStateObject
    {
        public Socket workSocket = null;
        public byte[] LenBuffer = new byte[4];
    }

    public class CTMessage
    {
        public string Id { get; set; }
        public string Command { get; set; }
        public string FileName { get; set; }
        public string Status { get; set; }
        public string Data { get; set; }
        public Int32 hWnd;

        public CTMessage()
        {
            Id = "";
            Command = "";
            FileName = "";
            Status = "";
            Data = "";
            hWnd = 0;
        }
    }
}